<?php

declare(strict_types=1);

namespace Drupal\entity_id_manager;

use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Query\Update;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Entity\ContentEntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\Sql\SqlEntityStorageInterface;
use Drupal\Core\TypedData\DataReferenceDefinitionInterface;

/**
 * Changes entity ids by direct database manipulation using Drupal\Core\Database\Connection
 * You should know that such manipulations remain unnoticed by Drupals integrity architecture.
 * No timestamps are updated, nor hooks, Events or pre- and post-callbacks are executed.
 * 
 * For Drupal, edits made using this service 'appear' like a swapped database with differenct entities and ids.
 * -> Clear your site cache (drush cache:clear) afterwards is mandatory!
 */
class EntityIdManager
{
  public function __construct(
    private readonly EntityTypeManagerInterface $entityTypeManager,
    private readonly EntityFieldManagerInterface $entityFieldManager,
    private readonly ModuleHandlerInterface $moduleHandler,
    private readonly Connection $connection
  ) {
  }

  /**
   * A pure SQL approach would use a recursive Common Table Expression (CTE)
   * 
   * @param string $entity_type_id
   * @param int $n_ids
   * @return int[]
   */
  public function getAvailableIds(string $entity_type_id, int $n_ids = 25): array
  {
    $storage = $this->entityTypeManager->getStorage($entity_type_id);
    $ids_used = $storage->getQuery()->accessCheck(FALSE)->execute();

    $ids = [];
    $id = 1;
    while (count($ids) < $n_ids) {
      if (!in_array($id, $ids_used)) {
        $ids[] = $id;
      }
      $id++;
    }

    return $ids;
  }

  /**
   * Creates a basic update query that changes an int value in a spcific table and column of the database
   * 
   * @param string $table
   * @param string $column
   * @param int $id_old
   * @param int $id_new
   * @return Update
   */
  private function update(string $table, string $column, int $id_old, int $id_new): Update
  {
    return $this->connection->update($table)
      ->condition($column, $id_old)
      ->fields([$column => $id_new]);
  }

  public function updateEntityTables(string $entity_type_id, int $id_old, int $id_new)
  {
    $storage = $this->entityTypeManager->getStorage($entity_type_id);
    $tableMapping = $storage->getTableMapping();

    // @see https://drupal.stackexchange.com/questions/310610/how-do-i-get-the-names-of-the-tables-where-field-data-is-stored
    foreach ($tableMapping->getTableNames() as $table) {

      // @todo Use better API methods to find the column names
      // f.e. $typed_data->getTargetIdentifier()

      $columns_all = $tableMapping->getAllColumns($table);
      $column = $this->entityTypeManager->getDefinition($entity_type_id)->getKey('id');
      if (in_array($column, $columns_all)) {
        $this->update($table, $column, $id_old, $id_new)
          ->execute();
      }


      // {$entity_type}__field_* tables name their id columns 'entity_id'
      $column = 'entity_id';
      if (in_array($column, $columns_all)) {
        $this->update($table, $column, $id_old, $id_new)
          ->execute();
      }
    }
  }

  protected function getEntityTypeFieldMap(): array
  {
    $map = [];
    foreach ($this->entityTypeManager->getDefinitions() as $entity_type_id => $entity_type) {
      if ($entity_type instanceof ContentEntityTypeInterface) {
        $map[$entity_type_id] = $this->entityFieldManager->getFieldStorageDefinitions($entity_type_id);
      }
    }

    return $map;
  }

  protected function getEntityReferenceTableMap(string $for_entity_type_id): array
  {
    $map = [];

    foreach ($this->getEntityTypeFieldMap() as $entity_type_id => $field_map) {
      foreach ($field_map as $field_id => $definition) {
        if (
          $data_reference = $definition->getPropertyDefinition('entity')
          and $data_reference instanceof DataReferenceDefinitionInterface
        ) {
          if ($data_reference->getTargetDefinition()->getEntityTypeId() == $for_entity_type_id) {
            $map[$entity_type_id][$field_id] = $definition;
          }
        }
      }
    }

    return $map;
  }

  // @todo add support for dynamic_entity_reference @FieldType
  public function updateEntityReferenceTables(string $entity_type_id, int $id_old, int $id_new)
  {
    foreach ($this->getEntityReferenceTableMap($entity_type_id) as $entity_type_id => $field_map) {
      $storage = $this->entityTypeManager->getStorage($entity_type_id);

      if ($storage instanceof SqlEntityStorageInterface) {
        $tableMapping = $storage->getTableMapping();

        foreach ($field_map as $field_id => $definition) {
          $column_name = $tableMapping->getFieldColumnName($definition, $definition->getMainPropertyName());
          foreach ($tableMapping->getAllFieldTableNames($field_id) as $table) {
            $this->update($table, $column_name, $id_old, $id_new)
              ->execute();
          }
        }
      }
    }
  }

  public function updateEntityReferenceModuleTables(string $module_name, string $entity_type_id, int $id_old, int $id_new): void
  {
    if (!$this->moduleHandler->moduleExists($module_name)) {
      return;
    }

    switch ($module_name) {
      case 'group':
        if ($entity_type_id == 'user') {
          $this->update('group_relationship_field_data', 'entity_id', $id_old, $id_new)
            ->condition('plugin_id', 'group_membership')
            ->execute();
        }
        break;
      case 'commerce_stock_local':
        $this->update('commerce_stock_transaction', 'entity_id', $id_old, $id_new)
          ->condition('entity_type', $entity_type_id)
          ->execute();
        break;
      case 'recently_read':
        $this->update('recently_read', 'entity_id', $id_old, $id_new)
          ->condition('type', $entity_type_id)
          ->execute();
        break;
    }
  }

  public function changeEntityId(string $entity_type_id, int $id_old, int $id_new): void
  {
    $this->updateEntityTables($entity_type_id, $id_old, $id_new);
    $this->updateEntityReferenceTables($entity_type_id, $id_old, $id_new);

    // Exceptional tables for some modules
    $this->updateEntityReferenceModuleTables('group', $entity_type_id, $id_old, $id_new);
    $this->updateEntityReferenceModuleTables('commerce_stock_local', $entity_type_id, $id_old, $id_new);
    $this->updateEntityReferenceModuleTables('recently_read', $entity_type_id, $id_old, $id_new);
  }
}
